@@ -1,6 +1,6 @@
-from typing import Any, Optional, Union, Sequence, Callable, Iterator, List
 from functools import partial, wraps
 from io import IOBase
+from typing import Any, Optional, Union, Sequence, Callable, Iterator, List
 import copy
 import hashlib
 import itertools
@@ -8,23 +8,27 @@ import json
 import multiprocessing as mp
 import os
 import pickle
+import re
 import shutil
 import string
 import subprocess
 import time
 import uuid
 import warnings
-import yaml
+import yaml  # type: ignore
+import sys
 
 # Using numpy type hints is complicated, so skip it for now.
 import numpy as np  # type: ignore
-import pandas as pd
+import pandas as pd  # type: ignore
 
 pj = os.path.join
 
-# pandas defaults
+# defaults, globals
 PANDAS_DEFAULT_ORIENT = "records"
 PANDAS_TIME_UNIT = "s"
+PSET_HASH_ALG = "sha1"
+GIT_ADD_ALL = "git add -A -v"
 
 
 # -----------------------------------------------------------------------------
@@ -70,29 +74,13 @@ def fullpath(path: str) -> str:
 
 
 def itr(func: Callable) -> Callable:
-    """Decorator which makes functions take a sequence of args or individual
-    args.
-
-    Examples
-    --------
-    >>> @itr
-    ... def func(args):
-    ...     for arg in args:
-    ...         print(arg)
-    >>> func([1,2,3])
-    1
-    2
-    3
-    >>> func(1,2,3)
-    1
-    2
-    3
-    """
-
     @wraps(func)
     def wrapper(*args):
+        # (arg1,)
         if len(args) == 1:
-            return func(args[0])
+            arg = args[0]
+            return func(arg if is_seq(arg) else [arg])
+        # (arg1,...,argN)
         else:
             return func(args)
 
@@ -135,15 +123,157 @@ def file_read(fn: str):
         return fd.read()
 
 
-def dict_hash(dct: dict, method="sha1"):
-    h = getattr(hashlib, method)()
-    h.update(json.dumps(dct, sort_keys=True).encode())
-    return h.hexdigest()
+class PsweepHashError(TypeError):
+    pass
+
+
+def pset_hash(
+    dct: dict,
+    method=PSET_HASH_ALG,
+    raise_error=True,
+    skip_special_cols=True,
+):
+    """Reproducible hash of a dict for usage in database (hash of a `pset`)."""
+
+    # We target "reproducible" hashes, i.e. not what Python's ``hash`` function
+    # does, for instance for two interpreter sessions::
+    #
+    #     $ python
+    #     >>> hash("12")
+    #     8013944793133897043
+    #
+    #     $ python
+    #     >>> hash("12")
+    #     4021864388667373027
+    #
+    # We only try to hash a dict if it is json-serializable. Things which
+    # aren't are also not hashable in a reproducible fashion w/o spending much
+    # more work to solve the problem. We can't hash, say, the pickled byte
+    # string of some object (e.g. ``hash(pickle.dumps(obj))``), b/c that may
+    # contain a ref to its memory location which is not what we're interested
+    # in. Similarly, also using ``repr`` is not reproducible::
+    #
+    #     >>> class Foo:
+    #     ...     pass
+    #
+    #     >>> repr(Foo())
+    #     '<__main__.Foo object at 0x7fcc68aa9d60>'
+    #     >>> repr(Foo())
+    #     '<__main__.Foo object at 0x7fcc732034c0>'
+    #
+    # even though for our purpose, we'd consider the two instances of ``Foo`` to
+    # be the same. Better be safe and return NaN in that case.
+    #
+    # The same observations have been also made elsewhere [1,2]. Esp. [2]
+    # points to [3] which in turn mentions joblib.hashing.hash(). It's code
+    # shows how complex the problem is. We may try joblib's solution in the
+    # future should this feature become so important that we need to require
+    # hashes to work or else fail. ATM we treat pset hashes on a best effort
+    # basis b/c we don't have strong use cases for them just yet.
+    #
+    # [1] https://death.andgravity.com/stable-hashing
+    # [2] https://ourpython.com/python/deterministic-recursive-hashing-in-python
+    # [3] https://stackoverflow.com/a/52175075
+    try:
+        if skip_special_cols:
+            keys = [x for x in dct.keys() if not x.startswith("_")]
+            _dct = {kk: dct[kk] for kk in keys}
+        else:
+            _dct = dct
+        h = hashlib.new(method)
+        h.update(json.dumps(_dct, sort_keys=True).encode())
+        return h.hexdigest()
+    except TypeError as ex:
+        if raise_error:
+            raise PsweepHashError(
+                f"Error in hash calculation of: {dct}"
+            ) from ex
+        else:
+            return np.nan
+
+
+def check_calc_dir(calc_dir: str, df: pd.DataFrame):
+    """Check calc dir for consistency with database.
+
+    Assuming dirs are named::
+
+        <calc_dir>/<pset_id1>
+        <calc_dir>/<pset_id2>
+        ...
+
+    check if we have matching dirs to ``_pset_id`` values in the
+    database.
+    """
+    # fmt: off
+    pset_ids_disk = set([
+        m.group()
+        for m in [
+            re.match(r"(([0-9a-z]+)-){4}([0-9a-z]+)", x)
+            for x in os.listdir(calc_dir)
+        ] if m is not None])
+    # fmt: on
+    pset_ids_db = set(df._pset_id.values)
+    return dict(
+        db_not_disk=pset_ids_db - pset_ids_disk,
+        disk_not_db=pset_ids_disk - pset_ids_db,
+    )
+
+
+def pickle_write(fn: str, obj):
+    with open(fn, "wb") as fd:
+        pickle.dump(obj, fd)
+
+
+def pickle_read(fn: str):
+    return pickle.load(open(fn, "rb"))
+
+
+# -----------------------------------------------------------------------------
+# git
+# -----------------------------------------------------------------------------
 
 
 def git_clean():
-    cmd = "git status --porcelain"
-    return system(cmd).stdout.decode() == ""
+    return system("git status --porcelain").stdout.decode() == ""
+
+
+def in_git_repo():
+    # fmt: off
+    return subprocess.run(
+        "git status",
+        check=False,
+        shell=True,
+        capture_output=True,
+        ).returncode == 0
+    # fmt: on
+
+
+def git_enter(use_git: bool, always_commit=False):
+    path = os.path.basename(fullpath(os.curdir))
+    if use_git:
+        if not in_git_repo():
+            if always_commit:
+                system(
+                    f"git init; {GIT_ADD_ALL}; git commit -m 'psweep: {path}: init'"
+                )
+            else:
+                raise Exception("no git repo here, create one first")
+        if not git_clean():
+            if always_commit:
+                print("dirty repo, adding all changes")
+                system(
+                    f"{GIT_ADD_ALL}; git commit -m 'psweep: {path}: local changes'"
+                )
+            else:
+                raise Exception("dirty repo, commit first")
+
+
+def git_exit(use_git: bool, df: pd.DataFrame):
+    path = os.path.basename(fullpath(os.curdir))
+    if use_git and (not git_clean()):
+        system(
+            f"{GIT_ADD_ALL}; git commit -m 'psweep: {path}: run_id={df._run_id.values[-1]}'"
+        )
 
 
 # -----------------------------------------------------------------------------
@@ -210,9 +340,34 @@ def df_read(fn: str, fmt="pickle", **kwds):
         raise Exception("unknown fmt: {}".format(fmt))
 
 
-def df_print(df: pd.DataFrame, index=False):
-    """Print DataFrame, by default without the index."""
-    print(df.to_string(index=index))
+def df_print(
+    df: pd.DataFrame,
+    index=False,
+    special_cols=False,
+    cols: Sequence[str] = None,
+):
+    """Print DataFrame, by default without the index and special fields such as
+    _pset_id.
+
+    Same logic as in bin/psweep-db2table but w/o tabulate support. Maybe
+    later.
+
+    Parameters
+    ----------
+    df : DataFrame
+    index : include DataFrame index
+    special_cols : include all special cols (_pset_id, ...)
+    cols : explicit list of cols, overrides special_cols
+    """
+    df2str = lambda df: df.to_string(index=index)
+    if cols is not None:
+        print(df2str(df[cols]))
+    else:
+        if special_cols:
+            print(df2str(df))
+        else:
+            _cols = set(x for x in df.columns if not x.startswith("_"))
+            print(df2str(df[_cols]))
 
 
 T = Union[pd.Series, pd.DataFrame, np.ndarray, List[bool]]
@@ -225,7 +380,7 @@ def df_filter_conds(df: pd.DataFrame, conds: Sequence[T]) -> pd.DataFrame:
 
     >>> df[conds[0] & conds[1] & conds[2] & ...]
 
-    but `conds` can be programatically generated while the expression above
+    but `conds` can be programmatically generated while the expression above
     would need to be changed by hand if `conds` changes.
 
     Parameters
@@ -286,12 +441,13 @@ def merge_dicts(args: Sequence[dict]):
     """Start with an empty dict and update with each arg dict
     left-to-right."""
     dct = {}
+    assert is_seq(args), f"input {args=} is no sequence"
     for entry in args:
+        assert isinstance(entry, dict), f"{entry=} is no dict"
         dct.update(entry)
     return dct
 
 
-@itr
 def itr2params(loops: Iterator[Any]):
     """Transform the (possibly nested) result of a loop over plists (or
     whatever has been used to create psets) to a proper list of psets
@@ -326,7 +482,12 @@ def itr2params(loops: Iterator[Any]):
     [{'a': 1, 'b': 77, 'c': 'const'},
      {'a': 2, 'b': 88, 'c': 'const'}]
     """
-    return [merge_dicts(flatten(entry)) for entry in loops]
+    ret = [merge_dicts(flatten(entry)) for entry in loops]
+    lens = list(map(len, ret))
+    assert (
+        len(np.unique(lens)) == 1
+    ), f"not all psets have same length {lens=}\n  {ret=}"
+    return ret
 
 
 @itr
@@ -350,9 +511,113 @@ def pgrid(plists):
     [{'a': 1, 'b': 77, 'c': 'const'},
      {'a': 2, 'b': 88, 'c': 'const'}]
     """
+    assert is_seq(plists), f"input {plists=} is no sequence"
     return itr2params(itertools.product(*plists))
 
 
+def filter_same_hash(params: Sequence[dict], **kwds):
+    """Reduce params to unique psets.
+
+    Parameters
+    ----------
+    params : Sequence of dicts
+    kwds : passed to pset_hash()
+
+    Returns
+    -------
+    params_filt
+    """
+    msk = np.unique(
+        [pset_hash(dct, **kwds) for dct in params], return_index=True
+    )[1]
+    return [params[ii] for ii in np.sort(msk)]
+
+
+def stargrid(
+    const: dict,
+    vary: Sequence[dict],
+    vary_labels: Sequence[str] = None,
+    vary_label_col: str = "_vary",
+    filter_dups=True,
+):
+    """
+    Helper to create a specific param sampling pattern.
+
+    Vary params in a "star" pattern (and not a full pgrid) around constant
+    values (middle of the "star").
+
+    When doing that, duplicate psets can occur. By default try to filter them
+    out (use filter_same_hash()) but ignore hash calculation errors and return
+    non-reduced params in that case. If you want to fail at hash errors, use
+
+    >>> filter_same_hash(stargrid(..., filter_dups=False), raise_error=True)
+
+    Examples
+    --------
+    >>> from psweep import psweep as ps
+    >>> const=dict(a=1, b=77, c=11)
+    >>> a=ps.plist("a", [1,2,3,4])
+    >>> b=ps.plist("b", [77,88,99])
+    >>> c=ps.plist("c", [11,22,33,44])
+
+    >>> ps.stargrid(const, vary=[a, b])
+    [{'a': 1, 'b': 77, 'c': 11},
+     {'a': 2, 'b': 77, 'c': 11},
+     {'a': 3, 'b': 77, 'c': 11},
+     {'a': 4, 'b': 77, 'c': 11},
+     {'a': 1, 'b': 88, 'c': 11},
+     {'a': 1, 'b': 99, 'c': 11}]
+
+    >>> ps.stargrid(const, vary=[a, b], filter_dups=False)
+    [{'a': 1, 'b': 77, 'c': 11},
+     {'a': 2, 'b': 77, 'c': 11},
+     {'a': 3, 'b': 77, 'c': 11},
+     {'a': 4, 'b': 77, 'c': 11},
+     {'a': 1, 'b': 77, 'c': 11},
+     {'a': 1, 'b': 88, 'c': 11},
+     {'a': 1, 'b': 99, 'c': 11}]
+
+    >>> ps.stargrid(const, vary=[a, b], vary_labels=["a", "b"])
+    [{'a': 1, 'b': 77, 'c': 11, '_vary': 'a'},
+     {'a': 2, 'b': 77, 'c': 11, '_vary': 'a'},
+     {'a': 3, 'b': 77, 'c': 11, '_vary': 'a'},
+     {'a': 4, 'b': 77, 'c': 11, '_vary': 'a'},
+     {'a': 1, 'b': 88, 'c': 11, '_vary': 'b'},
+     {'a': 1, 'b': 99, 'c': 11, '_vary': 'b'}]
+
+    >>> ps.stargrid(const, vary=[ps.itr2params(zip(a,c)),b], vary_labels=["a+c", "b"])
+    [{'a': 1, 'b': 77, 'c': 11, '_vary': 'a+c'},
+     {'a': 2, 'b': 77, 'c': 22, '_vary': 'a+c'},
+     {'a': 3, 'b': 77, 'c': 33, '_vary': 'a+c'},
+     {'a': 4, 'b': 77, 'c': 44, '_vary': 'a+c'},
+     {'a': 1, 'b': 88, 'c': 11, '_vary': 'b'},
+     {'a': 1, 'b': 99, 'c': 11, '_vary': 'b'}]
+    """
+    params = []
+    if vary_labels is not None:
+        assert len(vary_labels) == len(
+            vary
+        ), f"{vary_labels=} and {vary=} must have same length"
+    for ii, plist in enumerate(vary):
+        for dct in plist:
+            if vary_labels is not None:
+                label = {vary_label_col: vary_labels[ii]}
+                _dct = merge_dicts(dct, label)
+            else:
+                _dct = dct
+            params.append(merge_dicts(const, _dct))
+
+    if filter_dups:
+        try:
+            return filter_same_hash(
+                params, raise_error=True, skip_special_cols=True
+            )
+        except PsweepHashError:
+            return params
+    else:
+        return params
+
+
 # -----------------------------------------------------------------------------
 # run study
 # -----------------------------------------------------------------------------
@@ -380,17 +645,13 @@ def worker_wrapper(
     pset_id = str(uuid.uuid4())
     _pset = copy.deepcopy(pset)
     _time_utc = pd.Timestamp(time.time(), unit=PANDAS_TIME_UNIT)
-    hash_alg = "sha1"
-    try:
-        pset_hash = dict_hash(pset, hash_alg)
-    except TypeError:
-        pset_hash = np.nan
+    hash_alg = PSET_HASH_ALG
     update = {
         "_run_id": run_id,
         "_pset_id": pset_id,
         "_calc_dir": calc_dir,
         "_time_utc": _time_utc,
-        f"_pset_{hash_alg}": pset_hash,
+        f"_pset_{hash_alg}": pset_hash(pset, hash_alg, raise_error=False),
         "_pset_seq": pset_seq,
         "_run_seq": run_seq,
     }
@@ -423,6 +684,7 @@ def run_local(
     database_dir: Optional[str] = None,
     database_basename="database.pk",
     backup=False,
+    git=False,
 ) -> pd.DataFrame:
     """
     Parameters
@@ -441,8 +703,8 @@ def run_local(
         | int : use multiprocessing.Pool (even for ``poolsize=1``)
     save : bool
         save final DataFrame to ``<calc_dir>/database.pk`` (pickle format only)
-    tmpsave : bol
-        save results from this pset (the current DataFrame row) to
+    tmpsave : bool
+        save results from each `pset` in `params` (the current DataFrame row) to
         ``<calc_dir>/tmpsave/<run_id>/<pset_id>.pk`` (pickle format only)
     verbose : {bool, sequence of str}
         | bool : print the current DataFrame row
@@ -459,10 +721,16 @@ def run_local(
         ``<database_dir>/<database_basename>``
     backup : bool
         Make backup of ``<calc_dir>`` to ``<calc_dir>.bak_<timestamp>_run_id_<_run_id>``
+    git : bool
+        Use ``git`` to commit all files written and changed by the current run
+        (`_run_id`). Make sure to create a `.gitignore` manually before if
+        needed.
     """
 
     database_dir = calc_dir if database_dir is None else database_dir
 
+    git_enter(git)
+
     if simulate:
         calc_dir_sim = calc_dir + ".simulate"
         if os.path.exists(calc_dir_sim):
@@ -544,6 +812,8 @@ def run_local(
     if save:
         df_write(df, database_fn)
 
+    git_exit(git, df)
+
     return df
 
 
@@ -561,6 +831,8 @@ class Machine:
             ^^^^^^^^^^^^^^^^^^^^^^^^^------------- machine_dir
             templates/machines/<name>/jobscript
                                       ^^^^^^^^^--- template.basename
+
+
         """
         self.name = os.path.basename(os.path.normpath(machine_dir))
         self.template = FileTemplate(
@@ -587,7 +859,11 @@ class FileTemplate:
         return self.filename
 
     def fill(self, pset):
-        return string.Template(file_read(self.filename)).substitute(pset)
+        try:
+            return string.Template(file_read(self.filename)).substitute(pset)
+        except:
+            print(f"Failed to fill template: {self.filename}", file=sys.stderr)
+            raise
 
 
 def gather_calc_templates(calc_templ_dir):
@@ -619,11 +895,10 @@ def prep_batch(
     backup: bool = False,
 ) -> pd.DataFrame:
 
-    if git:
-        if not os.path.exists(".git"):
-            system("git init; git add -A; git commit -m 'psweep: init'")
-        if not git_clean():
-            raise Exception("dirty git repo")
+    """
+    Write files based on templates.
+    """
+    git_enter(git)
 
     calc_templates = gather_calc_templates(calc_templ_dir)
     machines = gather_machines(machine_templ_dir)
@@ -637,7 +912,7 @@ def prep_batch(
             )
         return {}
 
-    df = run_local(worker, params, calc_dir=calc_dir, backup=backup)
+    df = run_local(worker, params, calc_dir=calc_dir, backup=backup, git=False)
 
     msk_latest = df._run_seq == df._run_seq.values.max()
     msk_old = df._run_seq < df._run_seq.values.max()
@@ -659,9 +934,6 @@ def prep_batch(
             f"#!/bin/sh\n\nhere=$(pwd)\n{txt}\n",
         )
 
-    if git:
-        if not git_clean():
-            system(
-                f"git add -A; git commit -m 'psweep: run_id={df._run_id.values[-1]}'"
-            )
+    git_exit(git, df)
+
     return df
